# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"


class TestSummary:
    def __init__(self, system_desc, trades_desc):
        self.system_stats = system_desc
        self.trades_stats = trades_desc



from pyquanttrade.market import marketData

def test_get_data():
    data = marketData.get_data("AAPL", "2010-01-01", "2022-01-01")
    assert len(data) != 0

def test_get_data_lt():
    data = marketData.get_data_lt("AAPL", "2010-01-01")
    assert len(data) != 0
   

def test_get_data_and_visualize():
    marketData.get_data_and_visualize("AAPL", "2010-01-01", "2021-01-01")


# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import pandas as pd
import numpy as np

from pyquanttrade.engine.utils import (
    max_drawdown_ratio,
    max_drawdown_value,
    safe_div,
    safe_min,
    safe_sum,
    safe_mean,
)

class SystemSummaryStats:

    def __init__(self, data_stats):
        self.ticker_stats = {ticker:self._get_system_table_stats(data_stats.long_ticker[ticker], data_stats.short_ticker[ticker], data_stats.global_ticker[ticker], data_stats.tickers_capital[ticker]) for ticker in data_stats.tickers}
        all_capital = np.sum(list(data_stats.tickers_capital.values()))
        self.ticker_stats['all'] = self._get_system_table_stats(data_stats.long_all, data_stats.short_all, data_stats.global_all, all_capital)
    
    def  _generate_system_stats(self, system_results, capital):

        sim_years = system_results.shape[0] / 250
        profit = (
            system_results["Daily_capital"].tail(1)[0] - capital
        ) / capital
        global_results = [100 * profit]

        anual_net_profit_ratio = 100 * ((profit + 1) ** (1 / sim_years) - 1)
        global_results += [anual_net_profit_ratio]

        max_drawn_down = 100 * (max_drawdown_ratio(system_results["Daily_capital"]))
        global_results += [max_drawn_down]

        if max_drawn_down == 0:
            ajusted_annual_return = 0.0
        else:
            ajusted_annual_return = 100 * (anual_net_profit_ratio / max_drawn_down)
        global_results += [ajusted_annual_return]

        return global_results
    
    
    def _get_system_table_stats(self, long, short, system, capital):

        idx = [
            "Net profit%",
            "Annual Return %",
            "Max DrawDown",
            "Risk Adjusted Return %",
        ]


        data = {
            "All trades": self._generate_system_stats(system, capital),
            "Long trades": self._generate_system_stats(long, capital),
            "Short trades": self._generate_system_stats(short, capital),
        }

        return pd.DataFrame(data, index=idx).round(2)
    





# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"
import logging


class Trade:
    def __init__(
        self,
        order_type,
        ticker,
        day,
        price,
        num_actions=1,
        stop_loss=1,
        stop_loss_trailing=False,
        commission=0,
        slippage=0,
    ):

        self.type = order_type
        self.ticker = ticker
        self.open_day = day
        self.close_day = "None"
        self.close_price = "None"
        self.state = "OPENED"
        self.open_price = price
        self.init_capital = price * num_actions
        self.num_actions = num_actions
        self.profit = 0
        self.stop_loss = stop_loss
        self.up_price = self.open_price
        self.stop_loss_trailing = stop_loss_trailing
        self.commission = commission
        self.slippage = slippage
        self.num_days = 0

    def close(self, day, price):
        self.close_day = day
        if self.type == "long":
            self.close_price = max(price, (1 - self.stop_loss) * self.up_price)
            self.profit = (
                self.num_actions
                * (self.close_price - self.open_price - self.commission)
                - self.slippage
            )
        else:
            self.close_price = min(price, (1 + self.stop_loss) * self.up_price)
            self.profit = (
                self.num_actions
                * (self.open_price - self.close_price - self.commission)
                - self.slippage
            )

        self.state = "CLOSED"


class TradeList:
    def __init__(self, level):
        self.trade_index = float(0)
        self.trades_opened = {}
        self.trades_closed = {}
        logging.basicConfig(level=level)
        self.logger = logging.getLogger(__name__)

    def any_open_trade(self):
        return bool(self.trades_opened)

    def open_trade(
        self,
        ticker,
        order_type,
        day,
        price,
        num_actions=1,
        stop_loss=1,
        stop_loss_trailing=False,
        commission=0,
        slippage=0,
    ):
        self.trade_index += 1
        trade = Trade(
            order_type,
            ticker,
            day,
            price,
            num_actions,
            stop_loss,
            stop_loss_trailing,
            commission,
            slippage,
        )
        self.trades_opened[self.trade_index] = trade

    def close_open_trades(self, ticker, order_type, day, price):
        count = 0
        money = 0
        for key in list(self.trades_opened.keys()):
            v = self.trades_opened[key]
            if v.type == order_type and v.ticker == ticker:
                v.close(day, price)
                self.trades_closed[key] = v
                del self.trades_opened[key]
                money += v.init_capital + v.profit
                count += 1

        self.logger.info(str(count) + " trades close. Money free :" + str(money))
        return money

    def close_trade(self, k, day, price):
        v = self.trades_opened.pop(k)
        self.trades_closed[k] = v
        return v.init_capital + v.profit

    def get_current_profit(self, current_price, trade_type="all", ticker="all"):
        factor  = lambda t: 1.0 if t.type == "long" else -1.0
        open_price = (
            lambda trade, p: trade.num_actions
            * factor(trade) * (p - trade.open_price - trade.commission)
            - trade.slippage
        )

        if trade_type == "all":
            open_trade_prices = [
                (open_price(v, current_price), v.ticker) for k, v in self.trades_opened.items()
            ]
            close_trade_prices = [(v.profit, v.ticker) for k, v in self.trades_closed.items()]
        else:
            open_trade_prices = [
                (open_price(v, current_price), v.ticker)
                for k, v in self.trades_opened.items()
                if v.type == trade_type
            ]
            close_trade_prices = [
                (v.profit, v.ticker) for k, v in self.trades_closed.items() if v.type == trade_type
            ]
        if ticker != "all":
            open_trade_prices = [p for (p,t) in open_trade_prices if t == ticker]
            close_trade_prices = [p for (p,t) in close_trade_prices if t == ticker]
        else:
            open_trade_prices = [p for (p,t) in open_trade_prices]
            close_trade_prices = [p for (p,t) in close_trade_prices]
 
        return sum(open_trade_prices) + sum(close_trade_prices)

    def get_current_total_open(self, trade_type="all", ticker="all"):
        if trade_type == "all":
            open_trade_prices = [(v.init_capital, v.ticker) for k, v in self.trades_opened.items()]
            close_trade_prices = [(v.init_capital, v.ticker) for k, v in self.trades_closed.items()]
        else:
            open_trade_prices = [
                (v.init_capital, v.ticker)
                for k, v in self.trades_opened.items()
                if v.type == trade_type
            ]
            close_trade_prices = [
                (v.init_capital, v.ticker)
                for k, v in self.trades_closed.items()
                if v.type == trade_type
            ]

        if ticker != "all":
            open_trade_prices = [p for (p,t) in open_trade_prices if t == ticker]
            close_trade_prices = [p for (p,t) in close_trade_prices if t == ticker]
        else:
            open_trade_prices = [p for (p,t) in open_trade_prices]
            close_trade_prices = [p for (p,t) in close_trade_prices]

        return sum(open_trade_prices) + sum(close_trade_prices)

    def get_current_only_open(self, trade_type="all", ticker="all"):
        if trade_type == "all":
            open_trade_prices = [(v.init_capital, v.ticker) for k, v in self.trades_opened.items()]
        else:
            open_trade_prices = [
                (v.init_capital, v.ticker)
                for k, v in self.trades_opened.items()
                if v.type == trade_type
            ]
        if ticker != "all":
            open_trade_prices = [p for (p,t) in open_trade_prices if t == ticker]
        else:
            open_trade_prices = [p for (p,t) in open_trade_prices]

        return sum(open_trade_prices)

    def get_close_profit(self, trade_type="all", ticker="all"):
        if trade_type == "all":
            close_trade_prices = [(v.profit, v.ticker) for k, v in self.trades_closed.items()]
        else:
            close_trade_prices = [
                (v.profit, v.ticker) for k, v in self.trades_closed.items() if v.type == trade_type
            ]
        if ticker != "all":
            close_trade_prices = [p for (p,t) in close_trade_prices if t == ticker]
        else:
            close_trade_prices = [p for (p,t) in close_trade_prices]

        return sum(close_trade_prices)

    def get_close_trades_total_open(self, trade_type="all", ticker="all"):
        if trade_type == "all":
            close_trade_prices = [(v.init_capital, v.ticker) for k, v in self.trades_closed.items()]
        else:
            close_trade_prices = [
                (v.init_capital, v.ticker)
                for k, v in self.trades_closed.items()
                if v.type == trade_type
            ]
        if ticker != "all":
            close_trade_prices = [p for (p,t) in close_trade_prices if t == ticker]
        else:
            close_trade_prices = [p for (p,t) in close_trade_prices]

        return sum(close_trade_prices)

    def get_open_trades_total_open(self, trade_type="all", ticker="all"):
        if len(self.trades_opened) == 0:
            return 0

        if trade_type == "all":
            open_trade_prices = [(v.open_price, v.ticker) for k, v in self.trades_opened.items()]
        else:
            open_trade_prices = [
                (v.open_price, v.ticker)
                for k, v in self.trades_opened.items()
                if v.type == trade_type
            ]
        if ticker != "all":
            open_trade_prices = [p for (p,t) in open_trade_prices if t == ticker]
        else:
            open_trade_prices = [p for (p,t) in open_trade_prices]

        return sum(open_trade_prices)

    def get_close_profit_ratio(self, trade_type="all"):
        if trade_type == "all":
            profits = sum([trade.profit for key, trade in self.trades_closed.items()])
            init_capital = sum(
                [trade.open_price for key, trade in self.trades_closed.items()]
            )
            return profits / init_capital

        profits = sum(
            [
                trade.profit
                for key, trade in self.trades_closed.items()
                if trade.type == trade_type
            ]
        )
        init_capital = sum(
            [
                trade.open_price
                for key, trade in self.trades_closed.items()
                if trade.type == trade_type
            ]
        )
        return profits / init_capital

    def verify_stop_loss(self, day, ticker, min_price, max_price):
        money = 0
        for key in list(self.trades_opened.keys()):
            v = self.trades_opened[key]
            if v.ticker == ticker:
                if v.type == "long":
                    if (1 - v.stop_loss) * v.up_price > min_price:
                        self.logger.info(
                            str(day)
                            + " : "
                            + str(v.up_price)
                            + " - stop loss price: "
                            + str((1 - v.stop_loss) * v.up_price)
                            + " - min price: "
                            + str(min_price)
                        )
                        v.close(day, (1 - v.stop_loss) * v.up_price)
                        self.trades_closed[key] = v
                        money += v.init_capital + v.profit
                        del self.trades_opened[key]
                        self.logger.info(" trades close. Money free :" + str(money))
                    else:
                        # increment days counter:
                        v.num_days += 1

                    if v.stop_loss_trailing and max_price > v.up_price:
                        v.up_price = max_price
                else:
                    if (1 + v.stop_loss) * v.up_price < max_price:
                        self.logger.info(
                            str(day)
                            + " : "
                            + str(v.up_price)
                            + " - stop loss price: "
                            + str((1 + v.stop_loss) * v.up_price)
                            + " - max price: "
                            + str(max_price)
                        )
                        v.close(day, (1 + v.stop_loss) * v.up_price)
                        self.trades_closed[key] = v
                        money += v.init_capital + v.profit
                        del self.trades_opened[key]
                        self.logger.info(" trades close. Money free :" + str(money))

                    else:
                        # increment days counter:
                        v.num_days += 1

                    if v.stop_loss_trailing and min_price < v.up_price:
                        v.up_price = max_price
        return money

    def clone(self, ticker):
        trade_list = TradeList(logging.WARNING)
        trade_list.trades_opened = {index:trade for index,trade in self.trades_opened.items() if trade.ticker == ticker}
        trade_list.trades_closed = {index:trade for index,trade in self.trades_closed.items() if trade.ticker == ticker}
        return trade_list



"""
from trading.policy_analysis.searching_algos import grid_search, manhattan_distance
from trading.policy_analysis.walkforward import sequential_backtesting

from trading.policies_backtesting import policies


class std_policy(Policy):
    name = "test_policy"

    @classmethod
    def set_attribs(cls, param1, param2):
        cls.param1 = param1
        cls.param2 = param2
        cls.buffer = max(param1, param2)
        return cls

    long_stop_loss = 100
    short_stop_loss = 100
    long_stop_loss_trailling = 100
    short_stop_loss_trailling = 100

    @classmethod
    def buy_long_when(cls):
        return cross_of_values(moving_average(cls.param1), moving_average(cls.param2))

    @classmethod
    def close_long_when(cls):
        return cross_of_values(moving_average(cls.param2), moving_average(cls.param1))


def test_optimiser():
    param_set = {"param1": [5, 10], "param2": [20, 30]}
    grid_search(
        std_policy, param_set, ["2010-01-01", "2011-12-31"], "INTC", 1, [15, 20]
    )


def test_manhattan_distance():
    param_set_1 = [0, 2, 3]
    param_set_2 = [1, 3, 5]
    steps = [1, 0.5, 2]
    assert manhattan_distance(param_set_1, param_set_2, steps) == 4


def test_sequential_backtesting():
    param_set = {"param1": [5, 10], "param2": [20, 30]}
    (
        overall_backtest_results,
        rolling_backtest_results,
        iter_params,
    ) = sequential_backtesting(
        std_policy,
        'grid_search',
        param_set,
        ["2010-01-01", "2018-12-31"],
        "INTC",
        600,
        300,
        1,
        [15, 20],
    )
    assert iter_params

"""


from pyquanttrade.features import functions, indicators
from pyquanttrade.market import marketData

# Test two functions to check proper execution
def test_functions():
    data = marketData.get_data("AAPL", "2010-01-01", "2021-01-01")
    test_func_data_1 = functions.special_k()(data)
    test_func_data_2 = functions.triangular_weighting_ma()(data)
    assert (len(test_func_data_1) != 0) & (test_func_data_1 is not None)
    assert (len(test_func_data_2) != 0) & (test_func_data_2 is not None)

# Test two indicators to check proper execution
def test_indicators():
    data = marketData.get_data("AAPL", "2010-01-01", "2021-01-01")
    test_func_1 = functions.special_k()
    test_func_2 = functions.triangular_weighting_ma()
    cross_of_values_data = indicators.cross_of_values(test_func_1,test_func_2)('2010-12-10',None,None,data)
    greater_than_data = indicators.greater_than(test_func_1,test_func_2)('2010-12-10',None,None,data)
    assert  cross_of_values_data is not None
    assert greater_than_data is not None



from pyquanttrade.engine.commands import backtest, backtest_and_visualise
from tests.policy_battery import test_policy_1
from tests.policy_battery import test_policy_2

def test_backtest():
    backtest(test_policy_2,"TSLA", "2012-01-01", "2021-01-01")

def test_backtest_and_visualise():
    backtest_and_visualise(test_policy_2,"TSLA", "2012-01-01", "2021-01-01")

def test_backest_multiple_stocks():
    stock_list = ['XOM','FB']
    result,_ = backtest(test_policy_2, stock_list, "1999-01-01", "2022-01-01")
    summary_result = result.describe(True)
    assert summary_result

def test_x():
    from pandas import DataFrame
    from pyquanttrade.features.functions import moving_average
    from pyquanttrade.features.indicators import cross_of_values
    from pyquanttrade.policy import Policy
    import traceback

    def test_all_in_one(stock_list, excel_name):
    
        start_date = '1999-01-01'
        end_date = '2022-01-01'

        fast_sma = 6
        slow_sma = 70

        class TraillingStopLoss(Policy):

            name = "TraillingStopLoss"
            version = "1.0"

            plot_functions = ['moving_average_50_of_close', 'hello']
            # stop loss parameters:
            long_stop_loss = 0.05
            short_stop_loss = 100
            long_stop_loss_trailling = True
            short_stop_loss_trailling = False

            def description(self):
                return "Estrategia de tipo moving average crossover. Inversión en largo cuando la SMA_10 supera SMA_40. Inversión en corto cuando la SMA_10 cae por debajo de SMA_40. Las posiciones corto/largo se cierran cuando se abren las contrarias."

            @classmethod
            def buy_long_when(self):
                func1 = moving_average(fast_sma)
                func2 = moving_average(slow_sma)
                return cross_of_values(func2, func1)

            @classmethod
            def close_long_when(self):
                func1 = moving_average(fast_sma)
                func2 = moving_average(slow_sma)
                return cross_of_values(func1, func2)
            pass

        policy_list = [TraillingStopLoss]

        combinations = policy_list
        number_of_combinations = len(combinations)

        results = list()
        for number,policy in enumerate(combinations):
            try:
                result = backtest(policy, stock_list, start_date, end_date, commission=0.01, slippage_perc=0.01, time_buffer=70)
                results.append(result[0])
                print(f'Backtest calculated: {number+1}/{number_of_combinations}')
            except:
                print(f'Error while processing {policy.name}')
                print(traceback.format_exc())


        combinations_name = [[policy.name] for policy in combinations]
        results_df = DataFrame(combinations_name, columns=['policy'])

        results_describe = [result.describe(True) for result in results]
        result_describe_one_table = [result.trades_stats['all'].append(result.system_stats['all'])['All trades'] for result in results_describe]
        results_df = results_df.join(DataFrame(result_describe_one_table, index=range(len(result_describe_one_table))))
        results_df.to_excel(excel_name, sheet_name='All trades')


    stock_list = ['XOM', 'GOOGL','FB','CMCSA','CVX','SHEL','UPS','BA','UNP','AAPL','MSFT',
        'NVDA','JNJ','UNH','ABBV','NEE','SO','D','PG','KO','PEP','RIO','LYB','ECL','AMZN',
        'TSLA','WMT','V','JPM','PLD','AMT','CCI']
    test_all_in_one(stock_list, 'results_all.xlsx')
    stock_list = ['SO', 'GOOGL', 'AAPL', 'MSFT', 'NEE', 'AMZN', 'V', 'AMT']
    test_all_in_one(stock_list, 'results_high_recovery_ratio.xlsx')


# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"


def always_true():
    def return_function(when, ticket, trades, data):
        return True

    return return_function


def apply_all(criteria):
    def return_function(when, ticket, trades, data):
        return all([f(when, ticket, trades, data) for f in criteria])

    return return_function


def apply_any(criteria):
    def return_function(when, ticket, trades, data):
        return any([f(when, ticket, trades, data) for f in criteria])

    return return_function


def apply_not(criterium):
    def return_function(when, ticket, trades, data):
        return not criterium(when, ticket, trades, data)

    return return_function


def not_trade():
    def return_function(when, ticket, trades, data):
        return not trades.any_open_trade()

    return return_function


def eq(element, than):
    def return_function(when, ticket, trades, data):
        return element(when, ticket, trades, data) == than

    return return_function


def leq(element, than):
    def return_function(when, ticket, trades, data):
        return element(when, ticket, trades, data) <= than

    return return_function


def geq(element, than):
    def return_function(when, ticket, trades, data):
        return element(when, ticket, trades, data) >= than

    return return_function


def gt(element, than):
    def return_function(when, ticket, trades, data):
        return element(when, ticket, trades, data) > than

    return return_function


def lt(element, than):
    def return_function(when, ticket, trades, data):
        return element(when, ticket, trades, data) < than

    return return_function


def TF_indicator(func):
    def return_function(when, ticker, trades, data):
        data1 = func(data)
        column_name = data1.name
        return data[when, column_name]

    return return_function


def unit_indicator(func):
    def return_function(when, ticker, trades, data):
        data1 = func(data)
        column_name = f"unit_indicator_{data1.name}"
        if column_name not in data.columns:
            data[column_name] = data1 == 1
        return data[column_name][when]

    return return_function


def cross_of_values(func1, func2):
    def return_function(when, ticker, trades, data):
        data1 = func1(data)
        data2 = func2(data)
        column_name = f"diff_rows_{data1.name}_vs_{data2.name}"
        if column_name not in data.columns:
            data[column_name] = 0
            data.loc[data1 <= data2, [column_name]] = 1
            data[column_name] = data[column_name].diff()
            data[column_name].iloc[0] = 0
            data[column_name].loc[data[column_name] == 1] = True
            data[column_name].loc[data[column_name] != 1] = False
        return data[column_name][when]

    return return_function


def greater_than(func1, func2):
    def return_function(when, ticker, trades, data):
        data1 = func1(data)
        data2 = func2(data)
        column_name = f"{data1.name}_greater_than_{data2.name}"
        if column_name not in data.columns:
            data[column_name] = 0
            data.loc[data1 > data2, [column_name]] = 1
            data[column_name] = data[column_name].astype(bool)
        return data[column_name][when]

    return return_function


def lower_than(func1, func2):
    def return_function(when, ticker, trades, data):
        data1 = func1(data)
        data2 = func2(data)
        column_name = f"{data1.name}_lower_than_{data2.name}"
        if column_name not in data.columns:
            data[column_name] = 0
            data.loc[data1 < data2, [column_name]] = 1
            data[column_name] = data[column_name].astype(bool)
        return data[column_name][when]

    return return_function


def upwards_turn(func1):
    def return_function(when, ticker, trades, data):
        data1 = func1(data)
        column_name = f"upwards_turn_{data1.name}"
        if column_name not in data.columns:
            data[column_name] = 0
            data.loc[data1.diff(1) < 0, [column_name]] = -1
            data.loc[data1.diff(1) > 0, [column_name]] = 1
            data[column_name] = data[column_name].diff(1)
            data[column_name].iloc[0] = 0
            data.loc[data[column_name] > 0, [column_name]] = True
            data.loc[data[column_name] <= 0, [column_name]] = False
        return data[column_name][when]

    return return_function


def downwards_turn(func1):
    def return_function(when, ticker, trades, data):
        data1 = func1(data)
        column_name = f"downwards_turn_{data1.name}"
        if column_name not in data.columns:
            data[column_name] = 0
            data.loc[data1.diff(1) < 0, [column_name]] = -1
            data.loc[data1.diff(1) > 0, [column_name]] = 1
            data[column_name] = data[column_name].diff(1)
            data[column_name].iloc[0] = 0
            data.loc[data[column_name] >= 0, [column_name]] = False
            data.loc[data[column_name] < 0, [column_name]] = True
        return data[column_name][when]

    return return_function



# -*- coding: utf-8 -*-
"""Functions from market data"""

# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

from pyquanttrade import market
from pyquanttrade.market import marketData
from pyquanttrade.engine.trade import TradeList
from pyquanttrade.engine.test_result import TestResult
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import plotly.graph_objects as go
import logging

# TODO: end positions closes to compute final return
def backtest(
    policy,
    tickers,
    start_at,
    stop_at,
    capital=10000,
    commission=0,
    slippage_perc=0,
    level=logging.WARNING,
    time_buffer=250
):
    logging.basicConfig(level=level)
    if not isinstance(tickers, list):
        tickers = [tickers]

    trade_list = TradeList(level=level)
    first_day = (
        datetime.strptime(start_at, "%Y-%m-%d").date() - timedelta(days=time_buffer)
    ).isoformat()

    policies = {}
    data_sims = {}
    data_dict = {}
    policy_class = policy
    logging.info(policy_class)
    remaining_capital = {tick: capital for tick in tickers}
    for ticker in tickers:
        policies[ticker] = policy_class
        logging.info(policy_class.name)
        data = marketData.get_data(ticker, first_day, stop_at)
        data_sim = data.loc[data.index >= start_at]

        if not data_sims:
            result = TestResult(data_sim, ticker, capital)
        else:
            result.add_ticker(data_sim, ticker, capital)

        data_sims[ticker] = data_sim
        data_dict[ticker] = data

    def calculate_num_shares(capital, share_price):
        num_shares = int(capital / share_price)
        remaining = capital - num_shares * share_price
        return (num_shares, remaining)

    def execute_policy(policy, i, row, data, capital, ticker):
        last_day = data.iloc[data.index.get_loc(i) - 1]
        signals = policy.execute(
            str(last_day.name).split(" ")[0], data, trades=trade_list
        )
        remaining = capital
        for signal in signals:
            if signal == "Buy_long":

                logging.info(str(ticker) + " : " + str(i) + " -- " + signal)
                slip = slippage_perc * (row["high"] - row["low"])
                loss = policy_class.long_stop_loss
                trailling = policy_class.long_stop_loss_trailling
                shares, remaining = calculate_num_shares(remaining, row["open"])
                trade_list.open_trade(
                    ticker,
                    "long",
                    str(i),
                    row["open"],
                    shares,
                    loss,
                    trailling,
                    commission,
                    slip,
                )
                logging.info("num of shares buyed: " + str(shares))
                logging.info("cost of shares buyed: " + str(shares * row["open"]))

            if signal == "Close_long":

                logging.info(str(ticker) + " : " + str(i) + " -- " + signal)
                money = trade_list.close_open_trades(
                    ticker, "long", str(i), row["open"]
                )
                remaining += money

            if signal == "Sell_short":

                logging.info(str(ticker) + " : " + str(i) + " -- " + signal)
                slip = slippage_perc * (row["high"] - row["low"])
                loss = policy_class.short_stop_loss
                trailling = policy_class.short_stop_loss_trailling
                shares, remaining = calculate_num_shares(remaining, row["open"])
                trade_list.open_trade(
                    ticker,
                    "short",
                    str(i),
                    row["open"],
                    shares,
                    loss,
                    trailling,
                    commission,
                    slip,
                )

            if signal == "Close_short":

                logging.info(str(ticker) + " : " + str(i) + " -- " + signal)
                money = trade_list.close_open_trades(
                    ticker, "short", str(i), row["open"]
                )
                remaining += money
        return remaining

    for i, row in data_sims[tickers[0]].iterrows():
        for ticker in policies.keys():
            if i in data_sims[ticker].index:
                ticker_row = data_sims[ticker].loc[i]
                remaining_capital[ticker] = execute_policy(
                    policies[ticker],
                    i,
                    ticker_row,
                    data_dict[ticker],
                    remaining_capital[ticker], 
                    ticker
                )
                money = trade_list.verify_stop_loss(
                    str(i), ticker, ticker_row["low"], ticker_row["high"]
                )
                remaining_capital[ticker] += money

                result.update(i, trade_list, ticker)

    #Data returns only works with one ticker, for visualization purposes
    return result, data

def plot_activity_list(activity_list):
    
    fig = go.Figure()

    #Decoding signal list as separate lists
    activity_list['buy_long'] = None
    activity_list.loc[activity_list['signals']=='Buy_long','buy_long'] = activity_list.loc[activity_list['signals']=='Buy_long','close']
    activity_list['rebuy_long'] = None
    activity_list.loc[activity_list['signals']=='Rebuy_long','rebuy_long'] = activity_list.loc[activity_list['signals']=='Rebuy_long','close']
    activity_list['close_long'] = None
    activity_list.loc[activity_list['signals']=='Close_long','close_long'] = activity_list.loc[activity_list['signals']=='Close_long','close']
    activity_list['sell_short'] = None
    activity_list.loc[activity_list['signals']=='Sell_short','sell_short'] = activity_list.loc[activity_list['signals']=='Sell_short','close']
    activity_list['resell_short'] = None
    activity_list.loc[activity_list['signals']=='Resell_short','resell_short'] = activity_list.loc[activity_list['signals']=='Resell_short','close']
    activity_list['close_short'] = None
    activity_list.loc[activity_list['signals']=='Close_short','close_short'] = activity_list.loc[activity_list['signals']=='Close_short','close']

    fig.add_trace(go.Scatter(x=activity_list['close'].index, y=activity_list['close'], mode='lines', name='Price', line={'width':1,'color':'black'}))
    if not activity_list['buy_long'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['buy_long'].index, y=activity_list['buy_long'], mode='markers', name='Buy Long', marker ={'symbol':'arrow-up', 'size':9, 'color':'green'}))
    if not activity_list['rebuy_long'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['rebuy_long'].index, y=activity_list['rebuy_long'], mode='markers', name='ReBuy Long', marker ={'symbol':'triangle-ne', 'size':9, 'color':'green'}))
    if not activity_list['close_long'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['close_long'].index, y=activity_list['close_long'], mode='markers', name='Close Long', marker ={'symbol':'x', 'size':9, 'color':'green'}))
    if not activity_list['sell_short'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['sell_short'].index, y=activity_list['sell_short'], mode='markers', name='Sell Short', marker ={'symbol':'arrow-down', 'size':9, 'color':'red'}))
    if not activity_list['resell_short'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['resell_short'].index, y=activity_list['resell_short'], mode='markers', name='ReSell Short', marker ={'symbol':'triangle-se', 'size':9, 'color':'red'}))
    if not activity_list['close_short'].isnull().all(): 
        fig.add_trace(go.Scatter(x=activity_list['close_short'].index, y=activity_list['close_short'], mode='markers', name='Close Short', marker ={'symbol':'x', 'size':9, 'color':'red'}))
    
    return fig


def backtest_and_visualise(
    policy,
    ticker,
    start_at,
    stop_at,
    capital=10000,
    commission=0,
    slippage_perc=0,
    level=logging.WARNING,
    time_buffer=250,
):
    assert ticker is not list

    result, data = backtest(
        policy,
        ticker,
        start_at,
        stop_at,
        capital=10000,
        commission=0,
        slippage_perc=0,
        level=logging.WARNING,
        time_buffer=250)

    for _, trade in result.data.trade_list.trades_closed.items():
        if trade.type == 'long':
            data.loc[trade.open_day,'signals'] = 'Buy_long'
            data.loc[trade.close_day,'signals'] = 'Close_long'
        elif trade.type == 'short':
            data.loc[trade.open_day,'signals'] = 'Sell_short'
            data.loc[trade.close_day,'signals'] = 'Close_short'

    fig = plot_activity_list(data)

    for elem in policy.plot_functions:
        if elem in data.columns:
            fig.add_trace(go.Scatter(x = data.index, y = data[elem], mode='lines', name=elem, line={'width':1}))

    return result, fig


# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import pandas as pd
import numpy as np

from pyquanttrade.engine.utils import (
    max_drawdown_ratio,
    max_drawdown_value,
    safe_div,
    safe_min,
    safe_sum,
    safe_mean,
)

class DailyStats:
    def __init__(self, data_sim, ticker, capital):
        self.trade_list = None
        self.tickers = [ticker]
        self.tickers_capital = {}
        self.tickers_capital[ticker] = capital
        self.long_ticker = {}
        self.short_ticker = {}
        self.global_ticker = {}
        (l, s, t) = self._initDataframe(data_sim.index, ticker, data_sim, capital)
        self.long_ticker[ticker] = l
        self.short_ticker[ticker] = s
        self.global_ticker[ticker] = t

        self.long_all = None
        self.short_all = None
        self.global_all = None

        self.last_closed_profit = {}
        self.last_closed_inversion = {}
        self._last_trades_len = {}
        self.last_closed_profit[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}
        self.last_closed_inversion[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}
        self._last_trades_len[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}


    # Init dataframes stats values for a ticker data. 
    def _initDataframe(self, index_data, ticker, data_sim, capital):
        long_system_results = pd.DataFrame(index=index_data)
        short_system_results = pd.DataFrame(index=index_data)
        system_results = pd.DataFrame(index=index_data)

        # init  long trades
        shares = int(capital / data_sim.iloc[0]["close"])
        long_system_results["close"] = data_sim["close"]
        long_system_results["trade_profit"] = np.NaN
        long_system_results["trade_inversion"] = np.NaN
        long_system_results["trade_cum_inversion"] = np.NaN
        long_system_results["trade_cum_profit"] = np.NaN
        long_system_results["trade_close"] = np.NaN
        long_system_results["trade_sequence"] = np.NaN
        long_system_results["profit_sequence"] = 0.0
        long_system_results["Daily_cumulative"] = 1.0
        long_system_results["Daily_sequence"] = 1.0
        long_system_results["Daily_current_inversion"] = np.NaN
        long_system_results["last_closed_profit"] = np.NaN
        long_system_results["sum_open_price"] = np.NaN
        long_system_results["open_trades"] = 0.0
        long_system_results["Daily_capital"] = 0.0
        long_system_results["Daily_freeze"] = 0.0
        long_system_results["baseline"] = (
            long_system_results["close"] * shares
        )

        # init  short trades
        short_system_results["close"] = data_sim["close"]
        short_system_results["trade_profit"] = np.NaN
        short_system_results["trade_inversion"] = np.NaN
        short_system_results["trade_cum_inversion"] = np.NaN
        short_system_results["trade_cum_profit"] = np.NaN
        short_system_results["trade_close"] = np.NaN
        short_system_results["trade_sequence"] = np.NaN
        short_system_results["Daily_cumulative"] = 1.0
        short_system_results["Daily_sequence"] = 1.0
        short_system_results["open_trades"] = 0.0
        short_system_results["profit_sequence"] = 0.0
        short_system_results["Daily_current_inversion"] = np.NaN
        short_system_results["last_closed_profit"] = np.NaN
        short_system_results["sum_open_price"] = np.NaN
        short_system_results["Daily_capital"] = 0.0
        short_system_results["Daily_freeze"] = 0.0
        short_system_results["baseline"] = (
            short_system_results["close"] * shares
        )

        # init  All trades
        system_results["close"] = data_sim["close"]
        system_results["trade_profit"] = np.NaN
        system_results["trade_inversion"] = np.NaN
        system_results["trade_cum_inversion"] = np.NaN
        system_results["trade_cum_profit"] = np.NaN
        system_results["trade_close"] = np.NaN
        system_results["trade_sequence"] = np.NaN
        system_results["Daily_cumulative"] = 1.0
        system_results["Daily_sequence"] = 1.0
        system_results["open_trades"] = 0.0
        system_results["profit_sequence"] = 0.0
        system_results["Daily_current_inversion"] = np.NaN
        system_results["last_closed_profit"] = np.NaN
        system_results["sum_open_price"] = np.NaN
        system_results["Daily_capital"] = 0.0
        system_results["Daily_freeze"] = 0.0
        system_results["baseline"] = system_results["close"] * shares

        return (long_system_results, short_system_results, system_results)

    # Add new ticker for the stats
    def add_ticker(self, data_sim, ticker, capital):
        self.tickers += [ticker]
        self.tickers_capital[ticker] = capital
        (l, s, t) = self._initDataframe(data_sim.index, ticker, data_sim, capital)
        self.long_ticker[ticker] = l
        self.short_ticker[ticker] = s
        self.global_ticker[ticker] = t

        self.last_closed_profit[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}
        self.last_closed_inversion[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}
        self._last_trades_len[ticker] = {"long": 0.0, "short": 0.0, "all": 0.0}

    def _update(self, trade_type, i, row, trade_list, system_results, ticker):
        if trade_type == "all":
            trade_closed_len = (
                self._last_trades_len[ticker]["long"]
                + self._last_trades_len[ticker]["short"]
            )
        else:
            trade_closed_len = len(
                [
                    trade
                    for k, trade in trade_list.trades_closed.items()
                    if (trade.type == trade_type and trade.ticker == ticker)
                ]
            )

        if trade_closed_len > self._last_trades_len[ticker][trade_type]:

            self._last_trades_len[ticker][trade_type] = trade_closed_len
            total_closed_profit = trade_list.get_close_profit(trade_type, ticker)
            delta_closed_profit = (
                total_closed_profit - self.last_closed_profit[ticker][trade_type]
            )
            self.last_closed_profit[ticker][trade_type] = total_closed_profit

            total_closed_inversion = trade_list.get_close_trades_total_open(trade_type, ticker)
            
            delta_closed_inversion = (
                total_closed_inversion - self.last_closed_inversion[ticker][trade_type]
            )
            self.last_closed_inversion[ticker][trade_type] = total_closed_inversion

            system_results.loc[i, ["trade_profit"]] = delta_closed_profit
            system_results.loc[i, ["trade_inversion"]] = delta_closed_inversion
            system_results.loc[i, ["trade_cum_inversion"]] = total_closed_profit
            system_results.loc[i, ["trade_cum_profit"]] = total_closed_inversion

            system_results.loc[i, ["trade_close"]] = 1 + safe_div(
                delta_closed_profit, delta_closed_inversion
            )
            system_results.loc[i, ["trade_sequence"]] = 1 + safe_div(
                total_closed_profit, total_closed_inversion
            )

        current_inversion = trade_list.get_current_total_open(trade_type, ticker)
        system_results.loc[i, ["Daily_current_inversion"]] = current_inversion
        if current_inversion == 0:
            system_results.loc[i, ["Daily_sequence"]] = 1
        else:
            system_results.loc[i, ["Daily_sequence"]] = (
                1 + (trade_list.get_current_profit(row, trade_type, ticker)) / current_inversion
            )
            system_results.loc[i, ["profit_sequence"]] = trade_list.get_current_profit(
                row, trade_type, ticker
            )


        if trade_type == "all":
            len_open_trades = len(
                [trade for k, trade in trade_list.trades_opened.items() if trade.ticker == ticker]
            )
        else:
            len_open_trades = len(
                [
                    trade
                    for k, trade in trade_list.trades_opened.items()
                    if (trade.type == trade_type and trade.ticker == ticker)
                ]
            )

        if len_open_trades > 0:
            sum_open_price = trade_list.get_open_trades_total_open(trade_type, ticker)
            system_results.loc[i,["last_closed_profit"]] = self.last_closed_profit[ticker][trade_type]
            system_results.loc[i,["sum_open_price"]] = sum_open_price
            system_results.loc[i, ["Daily_cumulative"]] = (
                1
                + (
                    trade_list.get_current_profit(row, trade_type, ticker)
                    - self.last_closed_profit[ticker][trade_type]
                )
                / sum_open_price
            )
            system_results.loc[i, ["open_trades"]] = len_open_trades

        dayly_capital = self.tickers_capital[ticker] + trade_list.get_current_profit(row, trade_type, ticker)
        system_results.loc[i, ["Daily_capital"]] = dayly_capital
        system_results.loc[
            i, ["Daily_freeze"]
        ] = dayly_capital - trade_list.get_current_only_open(trade_type, ticker)

    def update(self, i, trade_list, ticker):
        self.trade_list = trade_list
        l = self.long_ticker[ticker]
        s = self.short_ticker[ticker]
        t = self.global_ticker[ticker]
        self._update("long", i, l.loc[i, "close"], trade_list, l, ticker)
        self._update("short", i, s.loc[i, "close"], trade_list, s, ticker)
        self._update("all", i, t.loc[i, "close"], trade_list, t, ticker)
    

    
    def generate_all_stats(self):
        self.long_all = self._join_all_tickers(list(self.long_ticker.values()))
        self.short_all = self._join_all_tickers(list(self.short_ticker.values()))
        self.global_all = self._join_all_tickers(list(self.global_ticker.values()))


    def _join_ticker_result(self, ticker_a_results, ticker_b_result):
        join_results = pd.DataFrame(index=ticker_a_results.index)
        join_results["close"] = ticker_a_results["close"] + ticker_b_result["close"]
        join_results["trade_profit"] = ticker_a_results["trade_profit"] + ticker_b_result["trade_profit"]
        join_results["trade_inversion"] = ticker_a_results["trade_inversion"] + ticker_b_result["trade_inversion"]
        join_results["trade_cum_inversion"] = ticker_a_results["trade_cum_inversion"] + ticker_b_result["trade_cum_inversion"]
        join_results["trade_cum_profit"] = ticker_a_results["trade_cum_profit"] + ticker_b_result["trade_cum_profit"]
        join_results["trade_close"] = join_results["trade_profit"].div(join_results["trade_inversion"]) + 1
        join_results["trade_sequence"] = join_results["trade_cum_profit"].div(join_results["trade_cum_inversion"]) + 1
        join_results["Daily_current_inversion"] = ticker_a_results["Daily_current_inversion"] + ticker_b_result["Daily_current_inversion"]
        join_results["profit_sequence"] = ticker_a_results["profit_sequence"].add(ticker_b_result["profit_sequence"],fill_value=ticker_b_result["profit_sequence"][0])
        join_results["Daily_sequence"] = join_results["profit_sequence"].div(join_results["Daily_current_inversion"]) + 1 
        join_results["last_closed_profit"] = ticker_a_results["last_closed_profit"] + ticker_b_result["last_closed_profit"]
        join_results["sum_open_price"] = ticker_a_results["sum_open_price"] + ticker_b_result["sum_open_price"]
        join_results["Daily_cumulative"] = (ticker_a_results["profit_sequence"] - ticker_a_results["last_closed_profit"]).div(ticker_b_result["sum_open_price"]) + 1
        join_results["open_trades"] = ticker_a_results["open_trades"] + ticker_b_result["open_trades"]
        join_results["Daily_capital"] = ticker_a_results["Daily_capital"].add(ticker_b_result["Daily_capital"],fill_value=ticker_b_result["Daily_capital"][0])
        join_results["Daily_freeze"] = ticker_a_results["Daily_freeze"] + ticker_b_result["Daily_freeze"]
        join_results["baseline"] = ticker_a_results["baseline"] + ticker_b_result["baseline"]

        return join_results

    def _join_all_tickers(self, tickers):
        if len(tickers) == 1:
            return tickers[0]
            
        join = self._join_ticker_result(tickers[0], tickers[1])

        for t in tickers[2:]:
            join = self._join_ticker_result(join, t)
        return join

